package com.sun.beans.finder;

import java.lang.reflect.Executable;
import com.sun.beans.TypeResolver;
import java.util.Arrays;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.lang.reflect.Modifier;
import sun.reflect.misc.ReflectUtil;
import com.sun.beans.util.Cache;
import java.lang.reflect.Method;

public final class MethodFinder extends AbstractFinder<Method>
{
    private static final Cache<Signature, Method> CACHE;
    private final String name;
    
    public static Method findMethod(Class<?> clazz, String str, Class<?>... array) throws NoSuchMethodException {
        if (str == null) {
            throw new IllegalArgumentException("Method name is not set");
        }
        PrimitiveWrapperMap.replacePrimitivesWithWrappers(array);
        Signature key = new Signature(clazz, str, array);
        try {
            Method method = (Method)MethodFinder.CACHE.get(key);
            return (method == null || ReflectUtil.isPackageAccessible(method.getDeclaringClass())) ? method : ((Method)MethodFinder.CACHE.create(key));
        }
        catch (SignatureException ex) {
            throw ex.toNoSuchMethodException("Method '" + str + "' is not found");
        }
    }
    
    public static Method findInstanceMethod(Class<?> clazz, String str, Class<?>... array) throws NoSuchMethodException {
        Method method = findMethod(clazz, str, array);
        if (Modifier.isStatic(method.getModifiers())) {
            throw new NoSuchMethodException("Method '" + str + "' is static");
        }
        return method;
    }
    
    public static Method findStaticMethod(Class<?> clazz, String str, Class<?>... array) throws NoSuchMethodException {
        Method method = findMethod(clazz, str, array);
        if (!Modifier.isStatic(method.getModifiers())) {
            throw new NoSuchMethodException("Method '" + str + "' is not static");
        }
        return method;
    }
    
    public static Method findAccessibleMethod(Method method) throws NoSuchMethodException {
        Class<Object> declaringClass = method.getDeclaringClass();
        if (Modifier.isPublic(declaringClass.getModifiers()) && ReflectUtil.isPackageAccessible(declaringClass)) {
            return method;
        }
        if (Modifier.isStatic(method.getModifiers())) {
            throw new NoSuchMethodException("Method '" + method.getName() + "' is not accessible");
        }
        Type[] genericInterfaces = declaringClass.getGenericInterfaces();
        int length = genericInterfaces.length;
        int i = 0;
        while (i < length) {
            Type type = genericInterfaces[i];
            try {
                return findAccessibleMethod(method, type);
            }
            catch (NoSuchMethodException ex) {
                ++i;
            }
        }
        return findAccessibleMethod(method, declaringClass.getGenericSuperclass());
    }
    
    private static Method findAccessibleMethod(Method method, Type type) throws NoSuchMethodException {
        String name = method.getName();
        Class[] parameterTypes = method.getParameterTypes();
        if ((type instanceof Class)) {
            return findAccessibleMethod(((Class)type).getMethod(name, parameterTypes));
        }
        if ((type instanceof ParameterizedType)) {
            ParameterizedType actual = (ParameterizedType)type;
            for (Method method2 : ((Class)actual.getRawType()).getMethods()) {
                if (method2.getName().equals(name)) {
                    Class[] parameterTypes2 = method2.getParameterTypes();
                    if (parameterTypes2.length == parameterTypes.length) {
                        if (Arrays.equals(parameterTypes, parameterTypes2)) {
                            return findAccessibleMethod(method2);
                        }
                        Type[] genericParameterTypes = method2.getGenericParameterTypes();
                        if (parameterTypes.length == genericParameterTypes.length && Arrays.equals(parameterTypes, TypeResolver.erase(TypeResolver.resolve(actual, genericParameterTypes)))) {
                            return findAccessibleMethod(method2);
                        }
                    }
                }
            }
        }
        throw new NoSuchMethodException("Method '" + name + "' is not accessible");
    }
    
    private MethodFinder(String name, Class<?>[] array) {
        super(array);
        this.name = name;
    }
    
    @Override
    protected boolean isValid(Method method) {
        return super.isValid(method) && method.getName().equals(this.name);
    }
    
    MethodFinder(String s, Class[] array, MethodFinder.1 cache) {
        this(s, array);
    }
    
    static {
        CACHE = new Cache<Signature, Method>(Cache.Kind.SOFT, Cache.Kind.SOFT) {
            @Override
            public Method create(Signature signature) {
                try {
                    return MethodFinder.findAccessibleMethod((Method)new MethodFinder(signature.getName(), signature.getArgs()).find(signature.getType().getMethods()));
                }
                catch (Exception ex) {
                    throw new SignatureException(ex);
                }
            }
        };
    }
}
